-- YGREC8/VHDL/ALU8/CLA8_ASIC.vhdl
-- created mer. janv.  1 20:37:18 CET 2020 by whygee@f-cpu.org
--
-- Released under the GNU AGPLv3 license
--
-- This Carry Look-Ahead tree is coupled with the ROP2 unit
-- to add, subtract and compare bytes with only the most basic
-- boolean gates.

Library ieee;
    use ieee.std_logic_1164.all;
Library work;
    use work.all;
Library SLV;
    use SLV.SLV_utils.all;
Library proasic3;
    use proasic3.all;

entity CLA8 is
  port(
    CarryIn,    -- fanout : 5
    CMPS : in SL;
    nP, nG, G, P, X : in  SLV8;
    CarryOut : out SL;
    SUM: out SLV8);
end CLA8;

architecture ASIC of CLA8 is
  Signal C : SLV8;
  Signal P2_0, P2_1, G2_0, G2_1n, C2, Carry, SignedOverflow : SL;
begin

  C(0) <= CarryIn;

  -- First CLA3 block:
  --inputs: CarryIn, G(0), nG(0), G(1), nG(1), nG(2), P(0), nP(0), P(1), nP(1), P(2), nP(2)
  cla1a: entity AO1B -- generic map(exclude=>"-X---X--")
                             port map(A=>CarryIn, B=>P(0),    C=>nG(0),                     Y=>C(1)); 
  cla1b: entity CLA3B(ASIC)  port map(A=>P(1),    B=>CarryIn, C=>P(0),  D=>G(0),  E=>nG(1), Y=>C(2));
  cla1c: entity NOR3         port map(A=>nP(0),   B=>nP(1),   C=>nP(2),                     Y=>P2_0);
  cla1d: entity CLA3B(ASIC)  port map(A=>P(2),    B=>G(0),    C=>P(1),  D=>G(1),  E=>nG(2), Y=>G2_0);
  cla1e: entity AO1          port map(A=>CarryIn, B=>P2_0,    C=>G2_0,                     Y=>C(3));

  -- Second CLA3 block:
  --inputs: C(3), G(3), nG(3), P(3), nP(3), G(4), nG(4), P(4), nP(4), G(5), nG(5), P(5), nP(5)

  -- layer 1:
  cla2a: entity NOR3         port map(A=>nP(3), B=>nP(4), C=>nP(5), Y=>P2_1);
  cla2b: entity CLA3BN(ASIC) port map(A=>nP(5), B=>nP(4), C=>nG(3), D=>nG(4), E=>G(5), Y=>G2_1n);

  -- layer 2:
  cla2c: entity AO1B -- generic map(exclude=>"-X---X--")
                             port map(A=>C(3), B=>P(3), C=>nG(3),                      Y=>C(4));
  cla2d: entity CLA3B(ASIC)  port map(A=>P(4), B=>P(3), C=>C(3),    D=>G(3), E=>nG(4), Y=>C(5));
  cla2e: entity AO1B -- generic map(exclude=>"-X---X--")
                             port map(A=>P2_1, B=>P(3),                      C=>G2_1n, Y=>C(6));
  cla2f: entity CLA3B(ASIC)  port map(A=>P2_1, B=>P2_0, C=>CarryIn, D=>G2_0, E=>G2_1n, Y=>C2);

  -- Last CLA3 block:
  cla3_a: entity AO1B -- generic map(exclude=>"-X---X--")
                             port map(A=>C2,   B=>P(6), C=>nG(6),                      Y=>C(7));
  cla3_b: entity CLA3B(ASIC) port map(A=>P(7), B=>P(6), C=>C2,      D=>G(6), E=>nG(7), Y=>Carry);
  sof:    entity AND2A       port map(A=>X(7), B=>CMPS, Y=>SignedOverflow);

  -- Final layer:
  GenSum: for I in 0 to 7 generate
    sumx:  entity XOR2 port map(A=>X(I), B=>C(I), Y=>SUM(I)); --   SUM <= X XOR C;
  end generate GenSum;
  sumxc:  entity XOR2 port map(A=>SignedOverflow, B=>Carry , Y=>CarryOut);

end ASIC;
