-- file YGREC8/ASM/ASM_tb.vhdl
-- created/rebooted mer. déc. 26 09:03:13 CET 2018 by Yann Guidon / ygdes.com
-- version lun. déc. 31 07:56:37 CET 2018
-- version mer. mars 13 17:55:48 CET 2019 : new instruction syntax
-- version dim. mars 17 04:03:26 CET 2019
-- Released under the GNU AGPLv3 license

Library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
Library work;
    use work.all;
    use work.ygrec8_asm.all;
Library SLV;
    use SLV.SLV_utils.all;
Library YGREC8;
    use YGREC8.ygrec8_def.all;

entity asm_tb is
end asm_tb;

architecture plip of asm_tb is
begin

  bench: process
    variable i, errors: integer := 0;
    variable result, expected, diff : SLV16;
    variable tmp, tmp2: string(1 to 18);

    procedure test_asm(t: string; mustpass: integer) is
      variable inst, inst2: SLV16;
      variable tmp: string(1 to 18);
      variable ctx: string(1 to 10) := "must pass ";
    begin
      if mustpass = 0 then 
        ctx:="must fail ";
      end if;
      inst := Assemble(t, ctx);
      if Is_X(inst) then
        report "'" & t & "' => " & SLV_to_bin(inst);
        assert mustpass=0 report "should have failed !" severity failure;
      else
        Disassemble(inst, "", tmp);
        report "'" & t & "' => " & SLV_to_bin(inst) & " => " & tmp;
        assert mustpass=1 report "should have passed !" severity failure;

        inst2 := Assemble(tmp, "retest");
        if inst2 /= inst then
          report "oupsie, doesn't reassemble to the same value : " & SLV_to_bin(inst2);
        end if;
      end if;
    end procedure;

  begin

    report CR & LF & CR & LF & "Stress the parser";
    test_asm("", 0);
    test_asm("$", 0);
    test_asm(" ; $", 0);
    test_asm(" ploooooooooooooooooooooooop   ", 0);
    test_asm(";", 0);

    report CR & LF & CR & LF & "Test special instructions";
    test_asm("NOP", 1);
    test_asm(" NOP ;$", 1);
    test_asm("NOP x", 0);
    test_asm("OVL", 0);
    test_asm("OVL ", 0);
    test_asm("OVL 1", 1);

    report CR & LF & CR & LF & "Test numbers as well";
    test_asm("OVL R1", 0);
    test_asm("OVL A", 0);
    test_asm("OVL 111111111b", 0);
    test_asm("OVL 11111111b", 1);
    test_asm("OVL -127", 1);
    test_asm("OVL -128", 1);
    test_asm("OVL -129", 0);
    test_asm("OVL 1", 1);
    test_asm("OVL 255", 1);
    test_asm("OVL 256", 0);
    test_asm("OVL -Ah", 1);
    test_asm("OVL -FFh", 0);

    report CR & LF & CR & LF & "Some unexpectedly tricky bits";
    test_asm("LDCL D1 D1", 1);
    test_asm("LDCL 1 D1 ", 1);
    test_asm("LDCH D1 D1", 1);
    test_asm("LDCH 1 D1 ", 1);


    report CR & LF & CR & LF & "Test registers too";
    test_asm(" IN 1", 0);
    test_asm(" IN D1", 0);
    test_asm(" IN D1 1", 0);
    test_asm(" IN 1 D1", 1);
    test_asm(" IN 1 A1", 1);
    test_asm(" IN 1 D2", 1);
    test_asm(" IN 1 A2", 1);
    test_asm(" IN 1 R1", 1);
    test_asm(" IN 1 R2", 1);
    test_asm(" IN 1 R3", 1);
    test_asm(" IN 1 PC", 1);
    test_asm("OUT D1 1", 0);
    test_asm("OUT 1 D1", 1);
    test_asm("OUT 1 A1", 1);
    test_asm("OUT 1 D2", 1);
    test_asm("OUT 1 A2", 1);
    test_asm("OUT 1 R1", 1);
    test_asm("OUT 1 R2", 1);
    test_asm("OUT 1 R3", 1);
    test_asm("OUT 1 PC", 1);

    report CR & LF & CR & LF & "Test extra parameters";
    test_asm("IN 123 A1 1", 0);

    report CR & LF & CR & LF & "Test that ADD instruction's RI4 form";
    test_asm("ADD -10 R1", 1); -- RI8
    test_asm("ADD  -9 R1", 1); -- RI8
    test_asm("ADD  -8 R1", 1); -- RI4
    test_asm("ADD  -7 R1", 1); -- RI4
    test_asm("ADD  -6 R1", 1); -- RI4
    test_asm("ADD  -5 R1", 1); -- RI4
    test_asm("ADD  -4 R1", 1); -- RI4
    test_asm("ADD  -3 R1", 1); -- RI4
    test_asm("ADD  -2 R1", 1); -- RI4
    test_asm("ADD  -1 R1", 1); -- RI4
    test_asm("ADD   0 R1", 1); -- should be RI8
    test_asm("ADD   1 R1", 1); -- RI4
    test_asm("ADD   2 R1", 1); -- RI4
    test_asm("ADD   3 R1", 1); -- RI4
    test_asm("ADD   4 R1", 1); -- RI4
    test_asm("ADD   5 R1", 1); -- RI4
    test_asm("ADD   6 R1", 1); -- RI4
    test_asm("ADD   7 R1", 1); -- RI4
    test_asm("ADD   8 R1", 1); -- RI8
    test_asm("ADD   9 R1", 1); -- RI8
    test_asm("ADD  10 R1", 1); -- RI8

    report CR & LF & CR & LF & "the SUB's RI4 form uses 0";
    test_asm("SUB -10 R1", 1);  -- RI8
    test_asm("SUB  -9 R1", 1);  -- RI8
    test_asm("SUB  -8 R1", 1);  -- RI4
    test_asm("SUB  -7 R1", 1);  -- RI4
    test_asm("SUB  -6 R1", 1);  -- RI4
    test_asm("SUB  -5 R1", 1);  -- RI4
    test_asm("SUB  -4 R1", 1);  -- RI4
    test_asm("SUB  -3 R1", 1);  -- RI4
    test_asm("SUB  -2 R1", 1);  -- RI4
    test_asm("SUB  -1 R1", 1);  -- RI4
    test_asm("SUB   0 R1", 1);  -- RI4, alias to NEG
    test_asm("SUB   1 R1", 1);  -- RI4
    test_asm("SUB   2 R1", 1);  -- RI4
    test_asm("SUB   3 R1", 1);  -- RI4
    test_asm("SUB   4 R1", 1);  -- RI4
    test_asm("SUB   5 R1", 1);  -- RI4
    test_asm("SUB   6 R1", 1);  -- RI4
    test_asm("SUB   7 R1", 1);  -- RI4
    test_asm("SUB   8 R1", 1);  -- RI8
    test_asm("SUB   9 R1", 1);  -- RI8
    test_asm("SUB  10 R1", 1);  -- RI8

    report CR & LF & CR & LF & "All the SND and SRI with all the opcodes";
    test_asm("XOR  D1 D1", 1);  -- RR
    test_asm(" OR  A1 A1", 1);  -- RR
    test_asm("AND  D2 D2", 1);  -- RR
    test_asm("ANDN A2 A2", 1);  -- RR
    test_asm("CMPU R1 R1", 1);  -- RR
    test_asm("CMPS R2 R2", 1);  -- RR
    test_asm("SUB  R3 R3", 1);  -- RR
    test_asm("ADD  PC PC", 1);  -- RR
    test_asm("SET  D1 D1", 1);  -- RR
    test_asm("CALL D1 D1", 1);  -- RR
    test_asm("SH   D1 D1", 1);  -- RR
    test_asm("SA   A1 A1", 1);  -- RR
    test_asm("RO   D2 D2", 1);  -- RR
    test_asm("RC   A2 A2", 1);  -- RR
    test_asm("LDCL R1 R1", 1);  -- RR
    test_asm("LDCH R2 R2", 1);  -- RR
    test_asm("IN   1  R3", 1);  -- RR
    test_asm("OUT  1  PC", 1);  -- RR

    report CR & LF & CR & LF & "All the conditions for RI4 and RR (including not allowed)" ;
    test_asm("ADD  D1 D1 NEVR", 1); 
    test_asm("ADD  D1 D1 IFN0", 1); 
    test_asm("ADD  D1 D1 IFNC", 1); 
    test_asm("ADD  D1 D1 IFN1", 1); 
    test_asm("ADD  D1 D1 IFN ", 1); 
    test_asm("ADD  D1 D1 IFN2", 1); 
    test_asm("ADD  D1 D1 IFNZ", 1); 
    test_asm("ADD  D1 D1 IFN3", 1); 
    test_asm("ADD  D1 D1 ALWS", 1); 
    test_asm("ADD  D1 D1 IF0 ", 1); 
    test_asm("ADD  D1 D1 IFC ", 1); 
    test_asm("ADD  D1 D1 IF1 ", 1); 
    test_asm("ADD  D1 D1 IFP ", 1); 
    test_asm("ADD  D1 D1 IF2 ", 1); 
    test_asm("ADD  D1 D1 IFZ ", 1); 
    test_asm("ADD  D1 D1 IF3 ", 1); 

    test_asm("ADD  1 D1 NEVR", 1); 
    test_asm("ADD  1 D1 IFN0", 0); 
    test_asm("ADD  1 D1 IFNC", 1); 
    test_asm("ADD  1 D1 IFN1", 0); 
    test_asm("ADD  1 D1 IFN ", 1); 
    test_asm("ADD  1 D1 IFN2", 0); 
    test_asm("ADD  1 D1 IFNZ", 1); 
    test_asm("ADD  1 D1 IFN3", 0); 
    test_asm("ADD  1 D1 ALWS", 1); 
    test_asm("ADD  1 D1 IF0 ", 0); 
    test_asm("ADD  1 D1 IFC ", 1); 
    test_asm("ADD  1 D1 IF1 ", 0); 
    test_asm("ADD  1 D1 IFP ", 1); 
    test_asm("ADD  1 D1 IF2 ", 0); 
    test_asm("ADD  1 D1 IFZ ", 1); 
    test_asm("ADD  1 D1 IF3 ", 0); 

    report CR & LF & CR & LF & "And now the pièce de résistance !" ;
    test_all_instruction: loop
      expected := std_logic_vector(to_unsigned(i,16));
      Disassemble(expected, SLV_to_bin(expected), tmp) ; 
      result := Assemble (tmp, SLV_to_bin(expected));
      Disassemble(result, SLV_to_bin(expected), tmp2) ;
      if result /= expected then
        diff := result xor expected;
        if not (    -- filter 4983 "errors" (not overlapping) :
              (diff(15 downto 12)="0000" and expected(15 downto 12)=Op_INV)  -- 4095 binary aliases of INV
           or (diff="0000100000000000" and tmp(1 to 3)="OVL")                -- 256 binary aliases of OVL
           or (diff="0000111000000000" and result(11 downto 9)="011")        -- 568 cases of  Imm8->Imm4
           or (expected(15 downto 12)=Op_ADD and diff(2 downto 0)="000"
                                          and diff(15 downto 7)="000011100") -- 64 cases of ADD Imm4+1
        ) then
          errors := errors+1;
          report integer'image(errors) & " : '" & SLV_to_bin(expected)
              & "' => '" & tmp & "' => '" & SLV_to_bin(result) & "'  "
              & tmp2 & "  XOR=" & SLV_to_bin(diff);
        end if;
      end if;
      i := i+1;
      exit when i>65535;
    end loop test_all_instruction;
    report CR & LF & CR & LF & "Errors : " & integer'image(errors);
    assert errors=0  report "Hey something's gone wrong here." severity failure;
    wait;
  end process;

end plip;
